@@ -14,7 +14,6 @@ See the example ``examples/association/proxied_association.py``.
 
 """
 import operator
-import weakref
 
 from .. import exc
 from .. import inspect
@@ -23,6 +22,7 @@ from .. import util
 from ..orm import collections
 from ..orm import interfaces
 from ..sql import or_
+from ..sql.operators import ColumnOperators
 
 
 def association_proxy(target_collection, attr, **kw):
@@ -105,6 +105,7 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
         proxy_factory=None,
         proxy_bulk_set=None,
         info=None,
+        cascade_scalar_deletes=False,
     ):
         """Construct a new :class:`.AssociationProxy`.
 
@@ -130,6 +131,15 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
           If you want to construct instances differently, supply a 'creator'
           function that takes arguments as above and returns instances.
 
+        :param cascade_scalar_deletes: when True, indicates that setting
+         the proxied value to ``None``, or deleting it via ``del``, should
+         also remove the source object.  Only applies to scalar attributes.
+         Normally, removing the proxied target will not remove the proxy
+         source, as this object may have other state that is still to be
+         kept.
+
+         .. versionadded:: 1.3
+
         :param getset_factory: Optional.  Proxied attribute access is
           automatically handled by routines that get and set values based on
           the `attr` argument for this proxy.
@@ -161,120 +171,93 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
         self.getset_factory = getset_factory
         self.proxy_factory = proxy_factory
         self.proxy_bulk_set = proxy_bulk_set
+        self.cascade_scalar_deletes = cascade_scalar_deletes
 
-        self.owning_class = None
         self.key = "_%s_%s_%s" % (
             type(self).__name__,
             target_collection,
             id(self),
         )
-        self.collection_class = None
         if info:
             self.info = info
 
-    @property
-    def remote_attr(self):
-        """The 'remote' :class:`.MapperProperty` referenced by this
-        :class:`.AssociationProxy`.
+    def __get__(self, obj, class_):
+        if class_ is None:
+            return self
+        inst = self._as_instance(class_, obj)
+        if inst:
+            return inst.get(obj)
 
-        .. versionadded:: 0.7.3
+        # obj has to be None here
+        # assert obj is None
 
-        See also:
+        return self
 
-        :attr:`.AssociationProxy.attr`
+    def __set__(self, obj, values):
+        class_ = type(obj)
+        return self._as_instance(class_, obj).set(obj, values)
+
+    def __delete__(self, obj):
+        class_ = type(obj)
+        return self._as_instance(class_, obj).delete(obj)
+
+    def for_class(self, class_, obj=None):
+        """Return the internal state local to a specific mapped class.
+
+        E.g., given a class ``User``::
+
+            class User(Base):
+                # ...
+
+                keywords = association_proxy('kws', 'keyword')
+
+        If we access this :class:`.AssociationProxy` from
+        :attr:`.Mapper.all_orm_descriptors`, and we want to view the
+        target class for this proxy as mapped by ``User``::
+
+            inspect(User).all_orm_descriptors["keywords"].for_class(User).target_class
+
+        This returns an instance of :class:`.AssociationProxyInstance` that
+        is specific to the ``User`` class.   The :class:`.AssociationProxy`
+        object remains agnostic of its parent class.
+
+        :param class_: the class that we are returning state for.
+
+        :param obj: optional, an instance of the class that is required
+         if the attribute refers to a polymorphic target, e.g. where we have
+         to look at the type of the actual destination object to get the
+         complete path.
+
+        .. versionadded:: 1.3 - :class:`.AssociationProxy` no longer stores
+           any state specific to a particular parent class; the state is now
+           stored in per-class :class:`.AssociationProxyInstance` objects.
 
-        :attr:`.AssociationProxy.local_attr`
 
         """
-        return getattr(self.target_class, self.value_attr)
+        return self._as_instance(class_, obj)
 
-    @property
-    def local_attr(self):
-        """The 'local' :class:`.MapperProperty` referenced by this
-        :class:`.AssociationProxy`.
+    def _as_instance(self, class_, obj):
+        try:
+            inst = class_.__dict__[self.key + "_inst"]
+        except KeyError:
+            owner = self._calc_owner(class_)
+            if owner is not None:
+                inst = AssociationProxyInstance.for_proxy(self, owner, obj)
+                setattr(class_, self.key + "_inst", inst)
+            else:
+                inst = None
 
-        .. versionadded:: 0.7.3
-
-        See also:
-
-        :attr:`.AssociationProxy.attr`
-
-        :attr:`.AssociationProxy.remote_attr`
-
-        """
-        return getattr(self.owning_class, self.target_collection)
-
-    @property
-    def attr(self):
-        """Return a tuple of ``(local_attr, remote_attr)``.
-
-        This attribute is convenient when specifying a join
-        using :meth:`.Query.join` across two relationships::
-
-            sess.query(Parent).join(*Parent.proxied.attr)
-
-        .. versionadded:: 0.7.3
-
-        See also:
-
-        :attr:`.AssociationProxy.local_attr`
-
-        :attr:`.AssociationProxy.remote_attr`
-
-        """
-        return (self.local_attr, self.remote_attr)
-
-    def _get_property(self):
-        owning_class = self.owning_class
-        if owning_class is None:
-            raise exc.InvalidRequestError(
-                "This association proxy has no mapped owning class; "
-                "can't locate a mapped property"
-            )
-        return orm.class_mapper(owning_class).get_property(
-            self.target_collection
-        )
-
-    @util.memoized_property
-    def target_class(self):
-        """The intermediary class handled by this :class:`.AssociationProxy`.
-
-        Intercepted append/set/assignment events will result
-        in the generation of new instances of this class.
-
-        """
-        return self._get_property().mapper.class_
-
-    @util.memoized_property
-    def scalar(self):
-        """Return ``True`` if this :class:`.AssociationProxy` proxies a scalar
-        relationship on the local side."""
-
-        scalar = not self._get_property().uselist
-        if scalar:
-            self._initialize_scalar_accessors()
-        return scalar
-
-    @util.memoized_property
-    def _value_is_scalar(self):
-        return (
-            not self._get_property()
-            .mapper.get_property(self.value_attr)
-            .uselist
-        )
-
-    @util.memoized_property
-    def _target_is_object(self):
-        return getattr(self.target_class, self.value_attr).impl.uses_objects
-
-    def _calc_owner(self, obj, class_):
-        if obj is not None and class_ is None:
-            target_cls = type(obj)
-        elif class_ is not None:
-            target_cls = class_
+        if inst is not None and not inst._is_canonical:
+            # the AssociationProxyInstance can't be generalized
+            # since the proxied attribute is not on the targeted
+            # class, only on subclasses of it, which might be
+            # different.  only return for the specific
+            # object's current value
+            return inst._non_canonical_get_for_object(obj)
         else:
-            return
+            return inst
 
+    def _calc_owner(self, target_cls):
         # we might be getting invoked for a subclass
         # that is not mapped yet, in some declarative situations.
         # save until we are mapped
@@ -285,69 +268,9 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
             # subclass, we can also scan through __mro__ to find a mapped
             # class, but instead just wait for us to be called again against a
             # mapped class normally.
-            return
-
-        # note we can get our real .key here too
-        owner = insp.mapper.class_manager._locate_owning_manager(self)
-        if owner is not None:
-            self.owning_class = owner.class_
+            return None
         else:
-            # the proxy is attached to a class that is not mapped
-            # (like a mixin), we are mapped, so, it's us.
-            self.owning_class = target_cls
-
-    def __get__(self, obj, class_):
-        if self.owning_class is None:
-            self._calc_owner(obj, class_)
-
-        if obj is None:
-            return self
-
-        if self.scalar:
-            target = getattr(obj, self.target_collection)
-            return self._scalar_get(target)
-        else:
-            try:
-                # If the owning instance is reborn (orm session resurrect,
-                # etc.), refresh the proxy cache.
-                creator_id, proxy = getattr(obj, self.key)
-                if id(obj) == creator_id:
-                    return proxy
-            except AttributeError:
-                pass
-            proxy = self._new(_lazy_collection(obj, self.target_collection))
-            setattr(obj, self.key, (id(obj), proxy))
-            return proxy
-
-    def __set__(self, obj, values):
-        if self.owning_class is None:
-            self._calc_owner(obj, None)
-
-        if self.scalar:
-            creator = self.creator and self.creator or self.target_class
-            target = getattr(obj, self.target_collection)
-            if target is None:
-                setattr(obj, self.target_collection, creator(values))
-            else:
-                self._scalar_set(target, values)
-        else:
-            proxy = self.__get__(obj, None)
-            if proxy is not values:
-                proxy.clear()
-                self._set(proxy, values)
-
-    def __delete__(self, obj):
-        if self.owning_class is None:
-            self._calc_owner(obj, None)
-
-        delattr(obj, self.key)
-
-    def _initialize_scalar_accessors(self):
-        if self.getset_factory:
-            get, set_ = self.getset_factory(None, self)
-        else:
-            get, set_ = self._default_getset(None)
-        self._scalar_get, self._scalar_set = get, set_
+            return insp.mapper.class_manager.class_
 
     def _default_getset(self, collection_class):
         attr = self.value_attr
@@ -368,31 +291,326 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
 
         return getter, setter
 
-    def _new(self, lazy_collection):
-        creator = self.creator and self.creator or self.target_class
-        self.collection_class = util.duck_type_collection(lazy_collection())
 
-        if self.proxy_factory:
-            return self.proxy_factory(
-                lazy_collection, creator, self.value_attr, self
+class AssociationProxyInstance(object):
+    """A per-class object that serves class- and object-specific results.
+
+    This is used by :class:`.AssociationProxy` when it is invoked
+    in terms of a specific class or instance of a class, i.e. when it is
+    used as a regular Python descriptor.
+
+    When referring to the :class:`.AssociationProxy` as a normal Python
+    descriptor, the :class:`.AssociationProxyInstance` is the object that
+    actually serves the information.   Under normal circumstances, its presence
+    is transparent::
+
+        >>> User.keywords.scalar
+        False
+
+    In the special case that the :class:`.AssociationProxy` object is being
+    accessed directly, in order to get an explicit handle to the
+    :class:`.AssociationProxyInstance`, use the
+    :meth:`.AssociationProxy.for_class` method::
+
+        proxy_state = inspect(User).all_orm_descriptors["keywords"].for_class(User)
+
+        # view if proxy object is scalar or not
+        >>> proxy_state.scalar
+        False
+
+    .. versionadded:: 1.3
+
+    """  # noqa
+
+    def __init__(self, parent, owning_class, target_class, value_attr):
+        self.parent = parent
+        self.key = parent.key
+        self.owning_class = owning_class
+        self.target_collection = parent.target_collection
+        self.collection_class = None
+        self.target_class = target_class
+        self.value_attr = value_attr
+
+    target_class = None
+    """The intermediary class handled by this
+    :class:`.AssociationProxyInstance`.
+
+    Intercepted append/set/assignment events will result
+    in the generation of new instances of this class.
+
+    """
+
+    @classmethod
+    def for_proxy(cls, parent, owning_class, parent_instance):
+        target_collection = parent.target_collection
+        value_attr = parent.value_attr
+        prop = orm.class_mapper(owning_class).get_property(target_collection)
+
+        # this was never asserted before but this should be made clear.
+        if not isinstance(prop, orm.RelationshipProperty):
+            raise NotImplementedError(
+                "association proxy to a non-relationship "
+                "intermediary is not supported"
             )
 
-        if self.getset_factory:
-            getter, setter = self.getset_factory(self.collection_class, self)
+        target_class = prop.mapper.class_
+
+        try:
+            target_assoc = cls._cls_unwrap_target_assoc_proxy(
+                target_class, value_attr
+            )
+        except AttributeError:
+            # the proxied attribute doesn't exist on the target class;
+            # return an "ambiguous" instance that will work on a per-object
+            # basis
+            return AmbiguousAssociationProxyInstance(
+                parent, owning_class, target_class, value_attr
+            )
         else:
-            getter, setter = self._default_getset(self.collection_class)
+            return cls._construct_for_assoc(
+                target_assoc, parent, owning_class, target_class, value_attr
+            )
 
-        if self.collection_class is list:
-            return _AssociationList(
-                lazy_collection, creator, getter, setter, self
+    @classmethod
+    def _construct_for_assoc(
+        cls, target_assoc, parent, owning_class, target_class, value_attr
+    ):
+        if target_assoc is not None:
+            return ObjectAssociationProxyInstance(
+                parent, owning_class, target_class, value_attr
             )
-        elif self.collection_class is dict:
-            return _AssociationDict(
-                lazy_collection, creator, getter, setter, self
+
+        attr = getattr(target_class, value_attr)
+        if attr._is_internal_proxy and not hasattr(attr, "impl"):
+            return AmbiguousAssociationProxyInstance(
+                parent, owning_class, target_class, value_attr
             )
-        elif self.collection_class is set:
-            return _AssociationSet(
-                lazy_collection, creator, getter, setter, self
+        is_object = attr.impl.uses_objects
+        if is_object:
+            return ObjectAssociationProxyInstance(
+                parent, owning_class, target_class, value_attr
+            )
+        else:
+            return ColumnAssociationProxyInstance(
+                parent, owning_class, target_class, value_attr
+            )
+
+    def _get_property(self):
+        return orm.class_mapper(self.owning_class).get_property(
+            self.target_collection
+        )
+
+    @property
+    def _comparator(self):
+        return self._get_property().comparator
+
+    @classmethod
+    def _cls_unwrap_target_assoc_proxy(cls, target_class, value_attr):
+        attr = getattr(target_class, value_attr)
+        if isinstance(attr, (AssociationProxy, AssociationProxyInstance)):
+            return attr
+        return None
+
+    @util.memoized_property
+    def _unwrap_target_assoc_proxy(self):
+        return self._cls_unwrap_target_assoc_proxy(
+            self.target_class, self.value_attr
+        )
+
+    @property
+    def remote_attr(self):
+        """The 'remote' :class:`.MapperProperty` referenced by this
+        :class:`.AssociationProxyInstance`.
+
+        ..seealso::
+
+            :attr:`.AssociationProxyInstance.attr`
+
+            :attr:`.AssociationProxyInstance.local_attr`
+
+        """
+        return getattr(self.target_class, self.value_attr)
+
+    @property
+    def local_attr(self):
+        """The 'local' :class:`.MapperProperty` referenced by this
+        :class:`.AssociationProxyInstance`.
+
+        .. seealso::
+
+            :attr:`.AssociationProxyInstance.attr`
+
+            :attr:`.AssociationProxyInstance.remote_attr`
+
+        """
+        return getattr(self.owning_class, self.target_collection)
+
+    @property
+    def attr(self):
+        """Return a tuple of ``(local_attr, remote_attr)``.
+
+        This attribute is convenient when specifying a join
+        using :meth:`.Query.join` across two relationships::
+
+            sess.query(Parent).join(*Parent.proxied.attr)
+
+        .. seealso::
+
+            :attr:`.AssociationProxyInstance.local_attr`
+
+            :attr:`.AssociationProxyInstance.remote_attr`
+
+        """
+        return (self.local_attr, self.remote_attr)
+
+    @util.memoized_property
+    def scalar(self):
+        """Return ``True`` if this :class:`.AssociationProxyInstance`
+        proxies a scalar relationship on the local side."""
+
+        scalar = not self._get_property().uselist
+        if scalar:
+            self._initialize_scalar_accessors()
+        return scalar
+
+    @util.memoized_property
+    def _value_is_scalar(self):
+        return (
+            not self._get_property()
+            .mapper.get_property(self.value_attr)
+            .uselist
+        )
+
+    @property
+    def _target_is_object(self):
+        raise NotImplementedError()
+
+    def _initialize_scalar_accessors(self):
+        if self.parent.getset_factory:
+            get, set_ = self.parent.getset_factory(None, self)
+        else:
+            get, set_ = self.parent._default_getset(None)
+        self._scalar_get, self._scalar_set = get, set_
+
+    def _default_getset(self, collection_class):
+        attr = self.value_attr
+        _getter = operator.attrgetter(attr)
+
+        def getter(target):
+            return _getter(target) if target is not None else None
+
+        if collection_class is dict:
+
+            def setter(o, k, v):
+                return setattr(o, attr, v)
+
+        else:
+
+            def setter(o, v):
+                return setattr(o, attr, v)
+
+        return getter, setter
+
+    @property
+    def info(self):
+        return self.parent.info
+
+    def get(self, obj):
+        if obj is None:
+            return self
+
+        if self.scalar:
+            target = getattr(obj, self.target_collection)
+            return self._scalar_get(target)
+        else:
+            try:
+                # If the owning instance is reborn (orm session resurrect,
+                # etc.), refresh the proxy cache.
+                creator_id, self_id, proxy = getattr(obj, self.key)
+            except AttributeError:
+                pass
+            else:
+                if id(obj) == creator_id and id(self) == self_id:
+                    assert self.collection_class is not None
+                    return proxy
+
+            self.collection_class, proxy = self._new(
+                _lazy_collection(obj, self.target_collection)
+            )
+            setattr(obj, self.key, (id(obj), id(self), proxy))
+            return proxy
+
+    def set(self, obj, values):
+        if self.scalar:
+            creator = (
+                self.parent.creator
+                if self.parent.creator
+                else self.target_class
+            )
+            target = getattr(obj, self.target_collection)
+            if target is None:
+                if values is None:
+                    return
+                setattr(obj, self.target_collection, creator(values))
+            else:
+                self._scalar_set(target, values)
+                if values is None and self.parent.cascade_scalar_deletes:
+                    setattr(obj, self.target_collection, None)
+        else:
+            proxy = self.get(obj)
+            assert self.collection_class is not None
+            if proxy is not values:
+                proxy._bulk_replace(self, values)
+
+    def delete(self, obj):
+        if self.owning_class is None:
+            self._calc_owner(obj, None)
+
+        if self.scalar:
+            target = getattr(obj, self.target_collection)
+            if target is not None:
+                delattr(target, self.value_attr)
+        delattr(obj, self.target_collection)
+
+    def _new(self, lazy_collection):
+        creator = (
+            self.parent.creator if self.parent.creator else self.target_class
+        )
+        collection_class = util.duck_type_collection(lazy_collection())
+
+        if self.parent.proxy_factory:
+            return (
+                collection_class,
+                self.parent.proxy_factory(
+                    lazy_collection, creator, self.value_attr, self
+                ),
+            )
+
+        if self.parent.getset_factory:
+            getter, setter = self.parent.getset_factory(collection_class, self)
+        else:
+            getter, setter = self.parent._default_getset(collection_class)
+
+        if collection_class is list:
+            return (
+                collection_class,
+                _AssociationList(
+                    lazy_collection, creator, getter, setter, self
+                ),
+            )
+        elif collection_class is dict:
+            return (
+                collection_class,
+                _AssociationDict(
+                    lazy_collection, creator, getter, setter, self
+                ),
+            )
+        elif collection_class is set:
+            return (
+                collection_class,
+                _AssociationSet(
+                    lazy_collection, creator, getter, setter, self
+                ),
             )
         else:
             raise exc.ArgumentError(
@@ -402,21 +620,9 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
                 % (self.collection_class.__name__, self.target_collection)
             )
 
-    def _inflate(self, proxy):
-        creator = self.creator and self.creator or self.target_class
-
-        if self.getset_factory:
-            getter, setter = self.getset_factory(self.collection_class, self)
-        else:
-            getter, setter = self._default_getset(self.collection_class)
-
-        proxy.creator = creator
-        proxy.getter = getter
-        proxy.setter = setter
-
     def _set(self, proxy, values):
-        if self.proxy_bulk_set:
-            self.proxy_bulk_set(proxy, values)
+        if self.parent.proxy_bulk_set:
+            self.parent.proxy_bulk_set(proxy, values)
         elif self.collection_class is list:
             proxy.extend(values)
         elif self.collection_class is dict:
@@ -429,16 +635,21 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
                 "collection_class implementation"
             )
 
-    @property
-    def _comparator(self):
-        return self._get_property().comparator
+    def _inflate(self, proxy):
+        creator = (
+            self.parent.creator and self.parent.creator or self.target_class
+        )
 
-    @util.memoized_property
-    def _unwrap_target_assoc_proxy(self):
-        attr = getattr(self.target_class, self.value_attr)
-        if isinstance(attr, AssociationProxy):
-            return attr
-        return None
+        if self.parent.getset_factory:
+            getter, setter = self.parent.getset_factory(
+                self.collection_class, self
+            )
+        else:
+            getter, setter = self.parent._default_getset(self.collection_class)
+
+        proxy.creator = creator
+        proxy.getter = getter
+        proxy.setter = setter
 
     def _criterion_exists(self, criterion=None, **kwargs):
         is_has = kwargs.pop("is_has", None)
@@ -509,6 +720,100 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
             criterion=criterion, is_has=True, **kwargs
         )
 
+
+class AmbiguousAssociationProxyInstance(AssociationProxyInstance):
+    """an :class:`.AssociationProxyInstance` where we cannot determine
+    the type of target object.
+    """
+
+    _is_canonical = False
+
+    def _ambiguous(self):
+        raise AttributeError(
+            "Association proxy %s.%s refers to an attribute '%s' that is not "
+            "directly mapped on class %s; therefore this operation cannot "
+            "proceed since we don't know what type of object is referred "
+            "towards"
+            % (
+                self.owning_class.__name__,
+                self.target_collection,
+                self.value_attr,
+                self.target_class,
+            )
+        )
+
+    def get(self, obj):
+        if obj is None:
+            self._ambiguous()
+        else:
+            return super(AmbiguousAssociationProxyInstance, self).get(obj)
+
+    def any(self, criterion=None, **kwargs):
+        self._ambiguous()
+
+    def has(self, criterion=None, **kwargs):
+        self._ambiguous()
+
+    @util.memoized_property
+    def _lookup_cache(self):
+        # mapping of <subclass>->AssociationProxyInstance.
+        # e.g. proxy is A-> A.b -> B -> B.b_attr, but B.b_attr doesn't exist;
+        # only B1(B) and B2(B) have "b_attr", keys in here would be B1, B2
+        return {}
+
+    def _non_canonical_get_for_object(self, parent_instance):
+        if parent_instance is not None:
+            actual_obj = getattr(parent_instance, self.target_collection)
+            if actual_obj is not None:
+                try:
+                    insp = inspect(actual_obj)
+                except exc.NoInspectionAvailable:
+                    pass
+                else:
+                    mapper = insp.mapper
+                    instance_class = mapper.class_
+                    if instance_class not in self._lookup_cache:
+                        self._populate_cache(instance_class, mapper)
+
+                    try:
+                        return self._lookup_cache[instance_class]
+                    except KeyError:
+                        pass
+
+        # no object or ambiguous object given, so return "self", which
+        # is a proxy with generally only instance-level functionality
+        return self
+
+    def _populate_cache(self, instance_class, mapper):
+        prop = orm.class_mapper(self.owning_class).get_property(
+            self.target_collection
+        )
+
+        if mapper.isa(prop.mapper):
+            target_class = instance_class
+            try:
+                target_assoc = self._cls_unwrap_target_assoc_proxy(
+                    target_class, self.value_attr
+                )
+            except AttributeError:
+                pass
+            else:
+                self._lookup_cache[instance_class] = self._construct_for_assoc(
+                    target_assoc,
+                    self.parent,
+                    self.owning_class,
+                    target_class,
+                    self.value_attr,
+                )
+
+
+class ObjectAssociationProxyInstance(AssociationProxyInstance):
+    """an :class:`.AssociationProxyInstance` that has an object as a target.
+    """
+
+    _target_is_object = True
+    _is_canonical = True
+
     def contains(self, obj):
         """Produce a proxied 'contains' expression using EXISTS.
 
@@ -536,7 +841,7 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
             )
         elif self._target_is_object and self.scalar and self._value_is_scalar:
             raise exc.InvalidRequestError(
-                "contains() doesn't apply to a scalar endpoint; use =="
+                "contains() doesn't apply to a scalar object endpoint; use =="
             )
         else:
 
@@ -560,32 +865,46 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
             getattr(self.target_class, self.value_attr) != obj
         )
 
-    def __repr__(self):
-        return "AssociationProxy(%r, %r)" % (
-            self.target_collection,
-            self.value_attr,
+
+class ColumnAssociationProxyInstance(
+    ColumnOperators, AssociationProxyInstance
+):
+    """an :class:`.AssociationProxyInstance` that has a database column as a
+    target.
+    """
+
+    _target_is_object = False
+    _is_canonical = True
+
+    def __eq__(self, other):
+        # special case "is None" to check for no related row as well
+        expr = self._criterion_exists(
+            self.remote_attr.operate(operator.eq, other)
+        )
+        if other is None:
+            return or_(expr, self._comparator == None)
+        else:
+            return expr
+
+    def operate(self, op, *other, **kwargs):
+        return self._criterion_exists(
+            self.remote_attr.operate(op, *other, **kwargs)
         )
 
 
 class _lazy_collection(object):
     def __init__(self, obj, target):
-        self.ref = weakref.ref(obj)
+        self.parent = obj
         self.target = target
 
     def __call__(self):
-        obj = self.ref()
-        if obj is None:
-            raise exc.InvalidRequestError(
-                "stale association proxy, parent object has gone out of "
-                "scope"
-            )
-        return getattr(obj, self.target)
+        return getattr(self.parent, self.target)
 
     def __getstate__(self):
-        return {"obj": self.ref(), "target": self.target}
+        return {"obj": self.parent, "target": self.target}
 
     def __setstate__(self, state):
-        self.ref = weakref.ref(state["obj"])
+        self.parent = state["obj"]
         self.target = state["target"]
 
 
@@ -639,6 +958,10 @@ class _AssociationCollection(object):
         self.lazy_collection = state["lazy_collection"]
         self.parent._inflate(self)
 
+    def _bulk_replace(self, assoc_proxy, values):
+        self.clear()
+        assoc_proxy._set(self, values)
+
 
 class _AssociationList(_AssociationCollection):
     """Generic, converting, list-to-list proxy."""
@@ -990,6 +1313,21 @@ class _AssociationDict(_AssociationCollection):
         for key, value in kw:
             self[key] = value
 
+    def _bulk_replace(self, assoc_proxy, values):
+        existing = set(self)
+        constants = existing.intersection(values or ())
+        additions = set(values or ()).difference(constants)
+        removals = existing.difference(constants)
+
+        for key, member in values.items() or ():
+            if key in additions:
+                self[key] = member
+            elif key in constants:
+                self[key] = member
+
+        for key in removals:
+            del self[key]
+
     def copy(self):
         return dict(self.items())
 
@@ -1074,6 +1412,24 @@ class _AssociationSet(_AssociationCollection):
         for value in other:
             self.add(value)
 
+    def _bulk_replace(self, assoc_proxy, values):
+        existing = set(self)
+        constants = existing.intersection(values or ())
+        additions = set(values or ()).difference(constants)
+        removals = existing.difference(constants)
+
+        appender = self.add
+        remover = self.remove
+
+        for member in values or ():
+            if member in additions:
+                appender(member)
+            elif member in constants:
+                appender(member)
+
+        for member in removals:
+            remover(member)
+
     def __ior__(self, other):
         if not collections._set_binops_check_strict(self, other):
             return NotImplemented
